"""
Python files to handle input and outputs
"""
from .atomic_program import AtomicProgram
from .software_configurations import SoftwareConfigurations
from .correction import Correction
from .minushalf_yaml import MinushalfYaml
from .input_file import InputFile
from .vtotal import Vtotal
from .make_minushalf_results import make_minushalf_results



"""
Class for correction input parameters in minushalf.yaml
"""
from shutil import Error
from minushalf.interfaces.minushalf_yaml_tags import MinushalfYamlTags
import loguru
from minushalf.data import CorrectionCode
from minushalf.interfaces import MinushalfYamlTags


class Correction(MinushalfYamlTags):
    """
    Set parameters and their default values
    """
    def __init__(self,
                 correction_code: str = CorrectionCode.get_default(),
                 potfiles_folder: str = "minushalf_potfiles",
                 amplitude: float = 1.0,
                 valence_cut_guess: list = None,
                 conduction_cut_guess: list = None,
                 tolerance: float = 0.01,
                 fractional_valence_treshold: float = 10,
                 fractional_conduction_treshold: float = 9,
                 cbm_characters: list = None,
                 vbm_characters: list = None,
                 overwrite_vbm: list = None,
                 overwrite_cbm: list = None,
                 inplace: bool = False,
                 divide_character: list = None) -> None:
        """
            Args:
                correction_code (str): Code for DFT -1/2 correction
                potfiles_folder (str): Potential files folder
                amplitude (float): Scaling factor in trimming function
                valence_cut_guess (float): Initial guess to valence correction
                conduction_cut_guess (float): Intitial guess to conduction correction
                tolerance (float): Absolute tolerance in the Nelder-Mead algorithm
                fractional_valence_treshold (float): Treshold for fractional valence correction
                fractional_conduction_treshold (float): Treshold for fractional conduction correction
                overwrite_vbm: Tag to overwrite vbm character
                overwrite_cbm: Tag to overwrite cbm character
                inplace: Realize calcualtions inplace
        """
        self.correction_code = correction_code
        self.potfiles_folder = potfiles_folder
        self.amplitude = amplitude
        self.valence_cut_guess = valence_cut_guess
        self.conduction_cut_guess = conduction_cut_guess
        self.tolerance = tolerance
        self.fractional_valence_treshold = fractional_valence_treshold
        self.fractional_conduction_treshold = fractional_conduction_treshold
        self.overwrite_cbm = overwrite_cbm
        self.overwrite_vbm = overwrite_vbm
        self.cbm_characters = cbm_characters
        self.vbm_characters = vbm_characters
        self.inplace = inplace
        self.divide_character = divide_character

    @property
    def divide_character(self) -> list:
        """
        Returns:
            Factor that divides the correction between atoms
        """
        return self._divide_character

    @divide_character.setter
    def divide_character(self, factors: int) -> None:
        """
        Set factor that divides the correction between atoms
        """
        try:
            if factors != None:
                for element in factors:
                    element[0] = element[0].capitalize()
                    element[1] = element[1].lower()
                    element[2] = int(element[2])
        except:
            loguru.logger.error("divide_character incorrectly specified")
            raise Error("divide_character incorrectly specified")

        self._divide_character = factors

    @property
    def vbm_characters(self) -> list:
        """
        Returns:
            Artificial character in vbm
        """
        return self._vbm_characters

    @vbm_characters.setter
    def vbm_characters(self, characters: list) -> None:
        """
        Set vbm_characters 
        """
        try:
            if characters != None:
                for element in characters:
                    element[0] = element[0].capitalize()
                    element[1] = element[1].lower()
                    element[2] = float(element[2])
        except:
            loguru.logger.error("repalce_vbm incorrectly specified")
            raise Error("repalce_vbm incorrectly specified")

        self._vbm_characters = characters

    @property
    def cbm_characters(self) -> list:
        """
        Returns:
            Artificial character in cbm
        """
        return self._cbm_characters

    @cbm_characters.setter
    def cbm_characters(self, characters: list) -> None:
        """
        Set cbm_characters 
        """

        try:
            if characters != None:
                for element in characters:
                    element[0] = element[0].capitalize()
                    element[1] = element[1].lower()
                    element[2] = float(element[2])
        except:
            loguru.logger.error("repalce_cbm incorrectly specified")
            raise Error("repalce_cbm incorrectly specified")

        self._cbm_characters = characters

    @property
    def valence_cut_guess(self) -> list:
        """
        Returns:
            CUT guess for nelder mead algorithm
        """
        return self._valence_cut_guess

    @valence_cut_guess.setter
    def valence_cut_guess(self, cut_guess: list) -> None:
        """
        Set valence_cut_guess
        """

        try:
            if cut_guess != None:
                for element in cut_guess:
                    element[0] = element[0].capitalize()
                    element[1] = element[1].lower()
                    element[2] = float(element[2])
        except:
            loguru.logger.error("valence_cut_guess incorrectly specified")
            raise Error("valence_cut_guess incorrectly specified")

        self._valence_cut_guess = cut_guess

    @property
    def conduction_cut_guess(self) -> list:
        """
        Returns:
            CUT guess for nelder mead algorithm
        """
        return self._conduction_cut_guess

    @conduction_cut_guess.setter
    def conduction_cut_guess(self, cut_guess: list) -> None:
        """
        Set conduction_cut_guess
        """

        try:
            if cut_guess != None:
                for element in cut_guess:
                    element[0] = element[0].capitalize()
                    element[1] = element[1].lower()
                    element[2] = float(element[2])
        except:
            loguru.logger.error("conduction_cut_guess incorrectly specified")
            raise Error("conduction_cut_guess incorrectly specified")

        self._conduction_cut_guess = cut_guess

    @property
    def correction_code(self) -> dict:
        """
        Returns:
            Code for DFT -1/2 correction
        """
        return self._correction_code

    @correction_code.setter
    def correction_code(self, code: str) -> None:
        """
        Set correction_code variable
        """
        available_correction_codes = CorrectionCode.to_list()
        is_code_available = any(test_code == code
                                for test_code in available_correction_codes)
        if not is_code_available:
            loguru.logger.error(
                "Invalid value for correction code in minushalf.yaml")
            raise ValueError("Invalid value for correction code")

        self._correction_code = code

    @property
    def overwrite_cbm(self) -> dict:
        """
        Returns:
            Tag to overwrite vbm character
        """
        return self._overwrite_cbm

    @overwrite_cbm.setter
    def overwrite_cbm(self, band_location: list) -> None:
        """
        Set overwirte_cbm variable
        """
        if not band_location:
            band_location = []
        self._overwrite_cbm = band_location

    @property
    def overwrite_vbm(self) -> dict:
        """
        Returns:
            Tag to overwrite vbm character
        """
        return self._overwrite_vbm

    @overwrite_vbm.setter
    def overwrite_vbm(self, band_location: list) -> None:
        """
        Set overwirte_vbm variable
        """
        if not band_location:
            band_location = []
        self._overwrite_vbm = band_location

    def to_list(self):
        """
        return list with the class variables
        """
        return list(self.__dict__.values())

    def to_dict(self):
        """
        Return dictionary with the class variables
        """
        parameters_dict = self.__dict__.copy()

        ## removing private variables
        parameters_dict["correction_code"] = parameters_dict.pop(
            "_correction_code", None)
        parameters_dict["overwrite_vbm"] = parameters_dict.pop(
            "_overwrite_vbm", None)
        parameters_dict["overwrite_cbm"] = parameters_dict.pop(
            "_overwrite_cbm", None)
        parameters_dict["vbm_characters"] = parameters_dict.pop(
            "_vbm_characters", None)
        parameters_dict["cbm_characters"] = parameters_dict.pop(
            "_cbm_characters", None)
        parameters_dict["divide_character"] = parameters_dict.pop(
            "_divide_character", None)
        parameters_dict["valence_cut_guess"] = parameters_dict.pop(
            "_valence_cut_guess", None)
        parameters_dict["conduction_cut_guess"] = parameters_dict.pop(
            "_conduction_cut_guess", None)

        return parameters_dict



"""
Class for atomic program software parameters in minushalf.yaml
"""
import loguru
from minushalf.data import Softwares
from minushalf.interfaces import MinushalfYamlTags


class SoftwareConfigurations(MinushalfYamlTags):
    """
    Set parameters and their default values
    """
    def __init__(
            self,
            command: list = None,
            software_name: str = Softwares.get_default(),
    ) -> None:
        """
            Args:
                command (list):  Command to perform first-principles calculations
                software_name (str): Name of software that performs first principles calculations
        """
        self.software_name = software_name
        self.command = command

    @property
    def command(self) -> dict:
        """
        Returns:
            Command to perform first-principles calculations
        """
        return self._command

    @command.setter
    def command(self, command_list: list) -> None:
        """
        Set command variable
        """
        if not command_list:
            software_commands = {Softwares.vasp.value: ['mpirun', 'vasp']}
            command_list = software_commands[self.software_name]

        self._command = command_list

    @property
    def software_name(self) -> str:
        """
        Returns:
            Name of the software used for ab initio calculations (VASP,...)
        """
        return self._software_name

    @software_name.setter
    def software_name(self, name: str) -> None:
        """
        Verify if the software name is a valid name

        Args:
            name (str): Name of the software
        """

        available_softwares = Softwares.to_list()
        is_software_avalilable = any(element.lower() == name.lower()
                                     for element in available_softwares)
        if not is_software_avalilable:
            loguru.logger.error("Parameter software is not filled correctly")
            raise ValueError("Parameter software is not filled correctly")

        self._software_name = name.upper()

    def to_list(self):
        """
        return list with the class variables
        """
        return [self._command]

    def to_dict(self):
        """
        Return dictionary with the class variables
        """
        return {"command": self._command}



"""
Class for atomic program input parameters in minushalf.yaml
"""
from minushalf.data import CalculationCode, ExchangeCorrelation
from minushalf.interfaces import MinushalfYamlTags


class AtomicProgram(MinushalfYamlTags):
    """
    Set parameters and their default values
    """
    def __init__(
        self,
        exchange_correlation_code: str = ExchangeCorrelation.get_default(),
        calculation_code: str = CalculationCode.get_default(),
        max_iterations: int = 100,
    ) -> None:
        """
            Args:
                exchange_correlation_code (str): Code for the exchange and correlation functional
                calculation_code (str):Calculation code for inp file (ae)
                max_iterations (int): Maximum number of iterations for atomic program
        """
        self.exchange_correlation_code = exchange_correlation_code
        self.calculation_code = calculation_code
        self.max_iterations = max_iterations

    def to_list(self):
        """
        return list with the class variables
        """
        return list(self.__dict__.values())

    def to_dict(self):
        """
        Return dictionary with the class variables
        """
        return self.__dict__



"""
Interface for minushalf.yaml
"""
from abc import ABC, abstractmethod


class MinushalfYaml(ABC):
    """
    Interface
    """
    @abstractmethod
    def get_correction_params(self) -> dict:
        """
        Get dictionary of correction parameters
        """

    @abstractmethod
    def get_atomic_program_params(self) -> dict:
        """
        Get dictionary of atomic program parameters
        """

    @abstractmethod
    def get_software_configurations_params(self) -> dict:
        """
        Get dictionary of software configurations parameters
        """

    @abstractmethod
    def get_software_name(self) -> str:
        """
        Returns the name of the software that runs first principles calculations
        """
    @abstractmethod
    def get_correction_code(self) -> str:
        """
        Returns the code used to identify the correction
        """
    @abstractmethod
    def get_overwrite_vbm(self) -> str:
        """
        Returns the parameter that overwrites vbm
        """

    @abstractmethod
    def get_overwrite_cbm(self) -> str:
        """
        Returns the parameter that overwrites cbm
        """

    @abstractmethod
    def get_vbm_characters(self) -> list:
        """
        Returns the parameter that cbm_characters
        """

    @abstractmethod
    def get_cbm_characters(self) -> list:
        """
        Returns the parameter vbm_characters
        """

    @abstractmethod
    def get_potential_folder(self) -> str:
        """
        Returns the potential folder name
        """

    @abstractmethod
    def get_amplitude(self) -> float:
        """
        Returns the amplitude
        """

    @abstractmethod
    def get_max_iterations(self) -> int:
        """
        Returns max iterations
        """

    @abstractmethod
    def get_exchange_corr_code(self) -> str:
        """
        Returns exchange correlation code
        """

    @abstractmethod
    def get_calculation_code(self) -> str:
        """
        Returns the calculation code
        """

    @abstractmethod
    def get_valence_cut_initial_guess(self) -> str:
        """
        Returns the valence cut initial guess
        """

    @abstractmethod
    def get_conduction_cut_initial_guess(self) -> str:
        """
        Returns the conduction cut initial guess
        """

    @abstractmethod
    def get_tolerance(self) -> float:
        """
        Returns the tolerance
        """

    @abstractmethod
    def get_inplace(self) -> bool:
        """
        Returns the inplace
        """

    @abstractmethod
    def get_divide_character(self) -> list:
        """
        Returns the divide characters
        """



"""
Reads Outcar file, an output of
VASP software
"""
import re
from collections import defaultdict


class Outcar():
    """
    Reads output informations stored
    in OUTCAR file
    """
    def __init__(self, filename: str):
        """
            Args:
                filename (str): name of the OUTCAR file in VASP

            Members:
                relative_distances (defaultdict(list)): An dictionary where
                the keys are the ion index given by VASP and the values are lists of tuples
                containing the index other ion and the relative distance
                respectively.
        """
        self.filename = filename
        self.relative_distances = self._get_distances()

    def nearest_neighbor_distance(self, ion_index: str) -> float:
        """
        Given the ion index, it returns the distance of the nearest neighbor
        to this ion

            Args:
                ion_index (str): The index of the ion given by VASP.

            Returns:
                nearest_neighbor_distance (float): The distance of the nearest neighbor.
        """
        distances = [dist for _, dist in self.relative_distances[ion_index]]
        return min(distances)

    def _get_ion_index(self, atoms_map: dict, target_symbol: str) -> int:
        """
        Given the symbol, returns the ion index
        """
        for index, symbol in atoms_map.items():
            if target_symbol == symbol:
                return index

    def _is_neighbor_equal(
        self,
        target_index: int,
        source_index: int,
        visited_neighbors: dict,
        atoms_map: dict,
        symbol: str,
    ) -> bool:
        """
        checks if elements have the same symbol and different indexes
        """
        different_index = target_index != source_index
        same_symbol = atoms_map[target_index] == symbol
        not_visited = not visited_neighbors[target_index]

        return different_index and same_symbol and not_visited

    def number_of_equal_neighbors(self, atoms_map: dict, symbol: str) -> int:
        """
        Given an map that links atoms symbols with it's index
        this function returns the number of neighbors of the atom with
        equal symbol but different indexes.

            Args:
                atoms_map (dict): Map the atoms index to their symbol.
                symbom (str): The symbol of the target atom.

            Returns:
                number_equal_neighbors (int): Returns the number of neighbors with
                                        same symbol but different indexes.
        """
        ion_index = self._get_ion_index(atoms_map, symbol)
        number_equal_neighbors = 0
        visited_neighbors = defaultdict(bool)  ## Already listed neighbors

        for index, _ in self.relative_distances[ion_index]:
            index = str(index)
            if self._is_neighbor_equal(index, ion_index,
                                       visited_neighbors, atoms_map, symbol):
                visited_neighbors[index] = True
                number_equal_neighbors += 1

        return number_equal_neighbors

    def _get_distances(self) -> defaultdict(list):
        """
            Returns:
                relative_distances (defaultdict(list)): An dictionary where
                the keys are the ion index given by VASP and the values are lists of tuples
                containing the index other ion and the relative distance
                respectively.
        """
        relative_distances = defaultdict(list)
        start_capture_regex = re.compile(
            r"\s*ion\s+position\s+nearest\s+neighbor\s+table")
        distances_line_regex = re.compile(
            r"\s+([0-9]).*(?=-)-\s+([0-9]\s+[0-9]*\.[0-9]+\s*)+")

        with open(self.filename) as outcar:

            ## Skip initial lines
            for line in outcar:
                if start_capture_regex.match(line):
                    break
            ## Read informations
            for line in outcar:

                if distances_line_regex.match(line):
                    ion_index = distances_line_regex.match(line).group(1)

                    ion_relative_distances_and_index = re.findall(
                        r"[0-9]\s+[0-9]*\.[0-9]+\s*",
                        line.split("-")[1])

                    for element in ion_relative_distances_and_index:
                        element = element.rstrip()
                        index, distance = int(element.split()[0]), float(
                            element.split()[1])
                        relative_distances[ion_index].append((index, distance))
                else:
                    break

        return relative_distances



"""
Extract the parameters for the correction
"""
import pandas as pd
from minushalf.interfaces import MinushalfYaml, SoftwaresAbstractFactory
from minushalf.utils import (
    get_simple_correction_indexes,
    get_fractionary_correction_indexes,
    CutInitialGuess,
)
from minushalf.data import (CutInitialGuessMethods)
from .band_structure import BandStructure
from .projection_to_df import projection_to_df


def _get_vbm_projection(factory: SoftwaresAbstractFactory) -> pd.DataFrame:
    """
    Returns vbm projection
    """
    band_structure = BandStructure.create(factory)
    vbm_projection = band_structure.vbm_projection()
    normalized_df = projection_to_df(vbm_projection)
    return normalized_df


def _get_cbm_projection(factory: SoftwaresAbstractFactory) -> pd.DataFrame:
    """
    Returns cbm projection
    """
    band_structure = BandStructure.create(factory)
    cbm_projection = band_structure.cbm_projection()
    normalized_df = projection_to_df(cbm_projection)
    return normalized_df


def _overwrite_band_projection(new_values: list,
                               band_projection: pd.DataFrame) -> pd.DataFrame:
    """
    Overwrite values in VBM or CBM band projection
        Args:
            new_values (list): Arguments passed in overwrite_vbm or overwrite_cbm.
            band_projection (pd.Dataframe): Dataframe with the value of projections
                                            of VBM or CBM.
        Returns:
            modified_band_projection (pd.Dataframe): Band projection data frame
                                                    with the values overwrited.
    """
    for case in new_values:
        atom = case[0].capitalize()
        orbital = case[1].lower()
        projection = int(case[2])
        band_projection[orbital][atom] = projection
    return band_projection


def _get_band_characters(band_location: list,
                         factory: SoftwaresAbstractFactory) -> pd.DataFrame:
    """
    Overwrite values in VBM or CBM band projection

        Args:

            band_location (List[int]): Kpoint and band number, respectively.

            factory (SoftwaresAbstractFactory): Software factory to create band structure class

        Returns:

            modified_band_projection (pd.Dataframe): Band projection data frame
                                                    with the values overwrited.

    """

    band_structure = BandStructure.create(factory)
    vbm_projection = band_structure.band_projection(*band_location)
    normalized_df = projection_to_df(vbm_projection)
    return normalized_df


def _get_valence_band_projection(minushalf_yaml: MinushalfYaml,
                                 software_factory: SoftwaresAbstractFactory):
    """
    Implements logic to return valence band projection
    """
    has_overwrite = bool(minushalf_yaml.get_overwrite_vbm())

    projection_df = None
    if has_overwrite:
        band_location = minushalf_yaml.get_overwrite_vbm()
        projection_df = _get_band_characters(band_location, software_factory)
    else:
        projection_df = _get_vbm_projection(software_factory)

    has_replace = bool(minushalf_yaml.get_vbm_characters())
    if has_replace:
        projection_df = _overwrite_band_projection(
            minushalf_yaml.get_vbm_characters(), projection_df)

    return projection_df


def _get_conduction_band_projection(
        minushalf_yaml: MinushalfYaml,
        software_factory: SoftwaresAbstractFactory):
    """
    Implements logic to return conduction band projection
    """
    has_overwrite = bool(minushalf_yaml.get_overwrite_cbm())

    projection_df = None
    if has_overwrite:
        band_location = minushalf_yaml.get_overwrite_cbm()
        projection_df = _get_band_characters(band_location, software_factory)
    else:
        projection_df = _get_cbm_projection(software_factory)

    has_replace = bool(minushalf_yaml.get_cbm_characters())
    if has_replace:
        projection_df = _overwrite_band_projection(
            minushalf_yaml.get_cbm_characters(), projection_df)

    return projection_df


def _guess_distance(symbol, software_factory):
    """
    Guess the distance
    """
    atoms_map = software_factory.get_atoms_map()
    ion_index = None
    for key, value in atoms_map.items():
        if value == symbol:
            ion_index = key
            break

    nearest_distance = software_factory.get_nearest_neighbor_distance(
        ion_index)
    cut_guesser = CutInitialGuess()
    return cut_guesser.guess(nearest_distance,
                             CutInitialGuessMethods.three_dimensions.value)


def _get_cut_initial_guess(initial_guess: list, correction_indexes: dict,
                           software_factory: SoftwaresAbstractFactory) -> dict:
    """
    Get cut initial guess parameter
    """
    if not initial_guess:
        initial_guess = []
    cut_guesses = {(e[0], e[1]): e[2] for e in initial_guess}
    for atom, orbitals in correction_indexes.items():
        for orbital in orbitals:
            if not (atom, orbital) in cut_guesses:
                cut_guesses[(atom, orbital)] = _guess_distance(
                    atom, software_factory)
    return cut_guesses


def _get_divide_character(divide_characters: list, correction_indexes: dict,
                          software_factory: SoftwaresAbstractFactory):
    """
    Get divide character param
    """
    if not divide_characters:
        divide_characters = []

    dividers = {(e[0], e[1]): e[2] for e in divide_characters}
    atoms_map = software_factory.get_atoms_map()
    for atom, orbitals in correction_indexes.items():
        for orbital in orbitals:
            if not (atom, orbital) in dividers:
                dividers[(
                    atom,
                    orbital)] = software_factory.get_number_of_equal_neighbors(
                        atoms_map=atoms_map, symbol=atom)
    return dividers


def _get_valence_correction_indexes(correction_code, band_projection):
    """
    Get the correction indexes needed to the corrrection
    """
    if "vf" in correction_code:
        return get_fractionary_correction_indexes(band_projection)
    else:
        return get_simple_correction_indexes(band_projection)


def _get_conduction_correction_indexes(correction_code, band_projection):
    """
    Get the correction indexes needed to the corrrection
    """
    if "cf" in correction_code:
        return get_fractionary_correction_indexes(band_projection)
    else:
        return get_simple_correction_indexes(band_projection)


def get_valence_correction_params(
    minushalf_yaml: MinushalfYaml,
    software_factory: SoftwaresAbstractFactory,
    **kwargs,
):
    """
    Returns the parameters for the valence correction
    """
    correction_code = minushalf_yaml.get_correction_code()
    params = kwargs
    params["software_factory"] = software_factory
    params["potential_filename"] = software_factory.get_potential_class(
    ).get_name()
    params["band_projection"] = _get_valence_band_projection(
        minushalf_yaml, software_factory)
    params["potential_folder"] = minushalf_yaml.get_potential_folder()
    params[
        "exchange_correlation_type"] = minushalf_yaml.get_exchange_corr_code()
    params["max_iterations"] = minushalf_yaml.get_max_iterations()
    params["calculation_code"] = minushalf_yaml.get_calculation_code()
    params["amplitude"] = minushalf_yaml.get_amplitude()
    params["tolerance"] = minushalf_yaml.get_tolerance()
    params["corrected_potfiles_folder"] = "corrected_valence_potfiles"
    params["correction_type"] = "valence"
    params["is_conduction"] = False
    params["inplace"] = minushalf_yaml.get_inplace()
    params["input_files"] = [
        "INCAR", "POSCAR", "KPOINTS", "CHGCAR"
    ] if params["inplace"] else ["INCAR", "POSCAR", "KPOINTS"]

    params["correction_indexes"] = _get_valence_correction_indexes(
        correction_code, params["band_projection"])
    params["cut_initial_guess"] = _get_cut_initial_guess(
        minushalf_yaml.get_valence_cut_initial_guess(),
        params["correction_indexes"], software_factory)

    params["divide_character"] = _get_divide_character(
        minushalf_yaml.get_divide_character(), params["correction_indexes"],
        software_factory)

    return params


def get_conduction_correction_params(
    minushalf_yaml: MinushalfYaml,
    software_factory: SoftwaresAbstractFactory,
    **kwargs,
):
    """
    Returns the parameters for the conduction correction
    """
    correction_code = minushalf_yaml.get_correction_code()
    params = kwargs
    params["software_factory"] = software_factory
    params["potential_filename"] = software_factory.get_potential_class(
    ).get_name()
    params["band_projection"] = _get_conduction_band_projection(
        minushalf_yaml, software_factory)
    params["potential_folder"] = "corrected_valence_potfiles"
    params[
        "exchange_correlation_type"] = minushalf_yaml.get_exchange_corr_code()
    params["max_iterations"] = minushalf_yaml.get_max_iterations()
    params["calculation_code"] = minushalf_yaml.get_calculation_code()
    params["amplitude"] = minushalf_yaml.get_amplitude()
    params["tolerance"] = minushalf_yaml.get_tolerance()
    params["corrected_potfiles_folder"] = "corrected_conduction_potfiles"
    params["correction_type"] = "conduction"
    params["is_conduction"] = True
    params["inplace"] = minushalf_yaml.get_inplace()
    params["input_files"] = [
        "INCAR", "POSCAR", "KPOINTS", "CHGCAR"
    ] if params["inplace"] else ["INCAR", "POSCAR", "KPOINTS"]
    params["correction_indexes"] = _get_conduction_correction_indexes(
        correction_code, params["band_projection"])
    params["cut_initial_guess"] = _get_cut_initial_guess(
        minushalf_yaml.get_valence_cut_initial_guess(),
        params["correction_indexes"], software_factory)

    params["divide_character"] = _get_divide_character(
        minushalf_yaml.get_divide_character(), params["correction_indexes"],
        software_factory)

    return params


"""
Interface for the handler class
"""
from abc import ABC, abstractmethod
from typing import Optional


class Handler(ABC):
    """
    The Handler interface declares a method for building the chain of handlers.
    It also declares a method for executing a request.
    """
    @abstractmethod
    def set_next(self, handler: any) -> any:
        """
        Set next handler
        """
        pass

    @abstractmethod
    def handle(self, request) -> Optional[any]:
        """
        Call action and pass to the next handler
        """
        pass
    
    @abstractmethod
    def action(self,request:any) -> any:
        """
        Performs the check or the process
        """
        pass



"""
Execute command
"""
import os
import sys
import shutil
from collections import OrderedDict
from typing import List
import numpy as np
import click
from loguru import logger
from minushalf.io import (MinushalfYaml, make_minushalf_results)
from minushalf.utils import (
    welcome_message,
    end_message,
    get_valence_correction_params,
    get_conduction_correction_params,
)
from minushalf.softwares import (Vasp)
from minushalf.corrections import (DFTCorrection)
from minushalf.data import (Softwares, CorrectionDefaultParams)
from minushalf.interfaces import (SoftwaresAbstractFactory)


def get_atoms_list(factory: SoftwaresAbstractFactory) -> list:
    """
    Returns atoms_list
    """
    atoms_map = factory.get_atoms_map()
    atoms = [atoms_map[key] for key in sorted(atoms_map)]
    return list(OrderedDict.fromkeys(atoms))


@click.command()
@click.option('--quiet', default=False, is_flag=True)
def execute(quiet: bool):
    """
    Uses the Nelder-Mead method to find
    the optimal values for the CUT(S) and,
    finally, find the corrected Gap value.
    This command uses external software to
    perform ab initio calculations, so it must
    be installed in order to perform the command.
    Check the docs for an list of the softwares supported
    by the CLI.


        Requires:


            minushalf.yaml : Parameters file. Check the docs
                            for a more detailed description.

            ab_initio_files: Files needed to perform the ab initio calculations.
                            They must be in the same directory as the input
                            file minushalf.yaml

            potential_folder: Folder with the potential files for each atom in
                            the crystal. The files must be named in the following pattern
                            ${POTENTIAL_FILE_NAME}.${LOWERCASE_CHEMICAL_SYMBOL}

        Returns:

            minushalf_results.dat : File that contains the optimal
                                values of the cuts and the final
                                value of the Gap.

            corrected_valence_potfiles: Potential files corrected with opti-mum valence cuts.

            corrected_conduction_potfiles: Potential files corrected with optimum conduction cuts.
    """
    welcome_message("minushalf")

    if quiet:
        logger.remove()
        logger.add(sys.stdout, level="ERROR")
    ## Read yaml file
    logger.info("Reading minushalf.yaml file")
    minushalf_yaml = MinushalfYaml.from_file()
    correction_factory_chooser = {Softwares.vasp.value: DFTCorrection}
    software_factory_chooser = {Softwares.vasp.value: Vasp()}

    software_name = minushalf_yaml.get_software_name()
    correction = correction_factory_chooser[software_name]
    software_factory = software_factory_chooser[software_name]

    ## Makes abinition calculation
    logger.info("Running ab initio calculations")
    software_configurations = minushalf_yaml.get_software_configurations_params(
    )
    runner = software_factory.get_runner(**software_configurations)
    runner.run()

    ## Makes root folder
    logger.info("Make potfiles folder")
    root_folder = "mkpotfiles"
    if os.path.exists(root_folder):
        shutil.rmtree(root_folder)
    os.mkdir(root_folder)

    ## get atoms list
    logger.info("Get atoms list")
    atoms = get_atoms_list(software_factory)

    ## amplitude logger
    if not np.isclose(minushalf_yaml.get_amplitude(),
                      CorrectionDefaultParams.amplitude.value):
        logger.warning(
            "Amplitude value is different from 1.0. This is not recommended unless you know exactly what you are doing."
        )

    valence_options = get_valence_correction_params(minushalf_yaml,
                                                    software_factory,
                                                    atoms=atoms,
                                                    runner=runner,
                                                    root_folder=root_folder)

    conduction_options = get_conduction_correction_params(
        minushalf_yaml,
        software_factory,
        atoms=atoms,
        runner=runner,
        root_folder=root_folder)

    correction_code = minushalf_yaml.get_correction_code()

    logger.info("Doing corrections")
    gap = None
    valence_cuts = None
    conduction_cuts = None

    if "v" in correction_code:
        valence_correction = correction(**valence_options)
        valence_cuts, valence_gap = valence_correction.execute()
        gap = valence_gap
        make_minushalf_results(valence_cuts=valence_cuts, gap=valence_gap)

    if "c" in correction_code:
        conduction_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_correction.execute()
        gap = conduction_gap

    make_minushalf_results(valence_cuts=valence_cuts,
                           gap=gap,
                           conduction_cuts=conduction_cuts)

    end_message()



"""
Parser for minushalf.yaml
"""
import yaml
import collections
import loguru
from minushalf.data import Softwares
from .software_configurations import SoftwareConfigurations
from .atomic_program import AtomicProgram
from .correction import Correction
from minushalf.interfaces import MinushalfYamlTags,MinushalfYaml


class MinushalfYaml(MinushalfYaml):
    """
    Class that parses the input
    for the execute command
    """
    def __init__(
        self,
        software_configurations: MinushalfYamlTags,
        atomic_program: MinushalfYamlTags,
        correction: MinushalfYamlTags,
    ):
        """
        Constructs a class for input in the execute command
        """
        self.software_configurations = software_configurations
        self.atomic_program = atomic_program
        self.correction = correction

    def get_correction_params(self) -> dict:
        """
        Get dictionary of correction parameters
        """
        return self.correction.to_dict()

    def get_atomic_program_params(self) -> dict:
        """
        Get dictionary of atomic program parameters
        """
        return self.atomic_program.to_dict()

    def get_software_configurations_params(self) -> dict:
        """
        Get dictionary of software configurations parameters
        """
        return self.software_configurations.to_dict()

    def get_software_name(self) -> str:
        """
        Returns the name of the software that runs first principles calculations
        """
        return self.software_configurations.software_name

    def get_command(self) -> list:
        """
        Returns the command that runs first principles calculations
        """
        return self.software_configurations.command

    def get_correction_code(self) -> list:
        """
        Returns the code used to identify the correction
        """
        return self.correction.correction_code

    def get_overwrite_vbm(self) -> list:
        """
        Returns the parameter that overwrites vbm
        """
        return self.correction.overwrite_vbm

    def get_overwrite_cbm(self) -> list:
        """
        Returns the parameter that overwrites cbm
        """
        return self.correction.overwrite_cbm

    def get_vbm_characters(self) -> list:
        """
        Returns the parameter that cbm_characters
        """
        return self.correction.vbm_characters

    def get_cbm_characters(self) -> list:
        """
        Returns the parameter vbm_characters
        """
        return self.correction.cbm_characters

    def get_potential_folder(self) -> str:
        """
        Returns the potential folder name
        """
        return self.correction.potfiles_folder

    def get_amplitude(self) -> float:
        """
        Returns the amplitude
        """
        return self.correction.amplitude

    def get_max_iterations(self) -> int:
        """
        Returns max iterations
        """
        return self.atomic_program.max_iterations

    def get_exchange_corr_code(self) -> str:
        """
        Returns exchange correlation code
        """
        return self.atomic_program.exchange_correlation_code

    def get_calculation_code(self) -> str:
        """
        Returns the calculation code
        """
        return self.atomic_program.calculation_code

    def get_valence_cut_initial_guess(self) -> str:
        """
        Returns the valence cut initial guess
        """
        return self.correction.valence_cut_guess

    def get_conduction_cut_initial_guess(self) -> str:
        """
        Returns the conduction cut initial guess
        """
        return self.correction.conduction_cut_guess

    def get_tolerance(self) -> float:
        """
        Returns the tolerance
        """
        return self.correction.tolerance

    def get_inplace(self) -> bool:
        """
        Returns the inplace
        """
        return self.correction.inplace

    def get_divide_character(self) -> list:
        """
        Returns the divide characters
        """
        return self.correction.divide_character

    @staticmethod
    def _read_yaml(filename: str) -> collections.defaultdict:
        """
        Read yaml file and export a dictionary
        """
        yaml_file = {}
        try:
            with open(filename, "r") as file:
                yaml_file = yaml.load(file, Loader=yaml.FullLoader)
        except FileNotFoundError:
            loguru.logger.warning(
                "File not found, default parameters will be used")

        return collections.defaultdict(dict, yaml_file)

    @staticmethod
    def _get_default_software_name(yaml_file: collections.defaultdict) -> str:
        """
        Name of the software that performs first priniple calculations
        """
        default = Softwares.vasp.value
        name = yaml_file["software"]

        software_name = (default if name == {} else name)
        return software_name

    @staticmethod
    def from_file(filename: str = "minushalf.yaml"):
        """
        Receives a file and catch all the parameters
        presents in the documentation
        """
        yaml_file = MinushalfYaml._read_yaml(filename)

        ## Get constructor parameters
        try:
            software_name = MinushalfYaml._get_default_software_name(yaml_file)

            software_configurations = SoftwareConfigurations(
                software_name=software_name, **yaml_file[software_name.lower()])
            correction = Correction(**yaml_file["correction"])
            atomic_program = AtomicProgram(**yaml_file["atomic_program"])
        except:
            loguru.logger.error("Invalid parameters in minushalf.yaml file")
            raise KeyError("Invalid parameters in minushalf.yaml file")

        return MinushalfYaml(
            software_configurations,
            atomic_program,
            correction,
        )


"""
Interface for minushalf.yaml tag classes
"""
from abc import ABC, abstractmethod


class MinushalfYamlTags(ABC):
    """
    Interface
    """
    @abstractmethod
    def to_list(self):
        """
        Get list of variables
        """

    @abstractmethod
    def to_dict(self):
        """
        Get dictionary of variables
        """
